@@ -1,11 +1,11 @@
 /*
  * acct.c	Accounting routines.
  *
+ *		Copyright 1994 Livingston Enterprises Inc
+ *		Copyright 1999-2001 Cistron Internet Services B.V.
  */
-char acct_sccsid[] =
-"@(#)acct.c	2.16 Copyright 1994 Livingston Enterprises Inc\n"
-"		     Copyright 1999,2000 Cistron Internet Services B.V.";
-
+char acct_rcsid[] =
+"$Id: acct.c,v 1.19 2001/11/26 21:51:58 miquels Exp $";
 
 #include	<sys/types.h>
 #include	<sys/socket.h>
@@ -32,6 +32,10 @@ static char trans[64] =
    "ABCDEFGHIJKLMNOPQRSTUVWXYZabcdefghijklmnopqrstuvwxyz0123456789+/";
 #define ENC(c) trans[c]
 
+#define NR_DETAIL_FN	8
+static char *detail_fn[NR_DETAIL_FN] = { "%N/detail" };
+static int detail_fn_count = 0;
+static char *detail_auth[] = { "Verified", "None", "Unverified" };
 static char porttypes[] = "ASITX";
 
 /*
@@ -60,6 +64,25 @@ static NAS_PORT *nas_port_find(UINT4 nasaddr, int port)
 	return cl;
 }
 
+/*
+ *	Lock / unlock file.
+ */
+static int lockfd(int fd, int lock)
+{
+	int		r;
+	int		op;
+
+#if defined(F_LOCK) && !defined(BSD)
+	op = lock ? (lock == 1 ? F_LOCK : F_TLOCK) : F_ULOCK;
+	r = lockf(fd, op, 0);
+#else
+	op = lock ? LOCK_EX : LOCK_UN;
+	if (lock == 2) lock |= LOCK_NB;
+	r = flock(fd, op);
+#endif
+	return r;
+}
+
 
 /*
  *	UUencode 4 bits base64. We use this to turn a 4 byte field
@@ -171,22 +194,19 @@ int radzap(UINT4 nasaddr, int port, char *user, time_t t, int dowtmp)
 
 	if (t == 0) time(&t);
 	if (dowtmp)
-		fp = fopen(RADWTMP, "a");
+		fp = fopen(radwtmp_path, "a");
 	else
 		fp = NULL;
 	netaddr = htonl(nasaddr);
 
-	if ((fd = open(RADUTMP, O_RDWR|O_CREAT, 0644)) >= 0) {
+	if ((fd = open(radutmp_path, O_RDWR|O_CREAT, 0644)) >= 0) {
 		int r;
 
 		/*
-		 *	Lock the utmp file, prefer lockf() over flock().
+		 *	Lock the utmp file.
 		 */
-#if defined(F_LOCK) && !defined(BSD)
-		(void)lockf(fd, F_LOCK, 0);
-#else
-		(void)flock(fd, LOCK_EX);
-#endif
+		lockfd(fd, 1);
+
 	 	/*
 		 *	Find the entry for this NAS / portno combination.
 		 */
@@ -227,7 +247,7 @@ int radzap(UINT4 nasaddr, int port, char *user, time_t t, int dowtmp)
 /*
  *	Store logins in the RADIUS utmp file.
  */
-int rad_accounting_new(AUTH_REQ *authreq, int dowtmp)
+int rad_accounting_radxtmp(AUTH_REQ *authreq, int dowtmp)
 {
 	struct radutmp	ut, u;
 	struct utmp	wt;
@@ -330,8 +350,10 @@ int rad_accounting_new(AUTH_REQ *authreq, int dowtmp)
 				break;
 			case PW_LOGIN_IP_HOST:
 			case PW_FRAMED_IP_ADDRESS:
-				framed_address = vp->lvalue;
-				ut.framed_address = htonl(vp->lvalue);
+				if (vp->lvalue != 0) {
+					framed_address = vp->lvalue;
+					ut.framed_address = htonl(vp->lvalue);
+				}
 				break;
 			case PW_FRAMED_PROTOCOL:
 				protocol = vp->lvalue;
@@ -340,7 +362,7 @@ int rad_accounting_new(AUTH_REQ *authreq, int dowtmp)
 				nas_address = vp->lvalue;
 				ut.nas_address = htonl(vp->lvalue);
 				break;
-			case PW_NAS_PORT_ID:
+			case PW_NAS_PORT:
 				ut.nas_port = vp->lvalue;
 				port_seen = 1;
 				break;
@@ -387,12 +409,14 @@ int rad_accounting_new(AUTH_REQ *authreq, int dowtmp)
 		ut.nas_address = htonl(nas_address);
 	}
 
-	if (protocol == PW_PPP)
-		ut.proto = 'P';
+	if (protocol < 0)
+		ut.proto = 'T';
 	else if (protocol == PW_SLIP)
 		ut.proto = 'S';
+	else if (protocol == PW_PPP)
+		ut.proto = 'P';
 	else
-		ut.proto = 'T';
+		ut.proto = 'P';
 	ut.time = t - ut.delay;
 	make_wtmp(&ut, &wt, status);
 
@@ -427,7 +451,7 @@ int rad_accounting_new(AUTH_REQ *authreq, int dowtmp)
 	 *	Perhaps we don't want to store this record into
 	 *	radutmp/radwtmp. We skip records:
 	 *
-	 *	- without a NAS-Port-Id (telnet / tcp access)
+	 *	- without a NAS-Port (telnet / tcp access)
 	 *	- with the username "!root" (console admin login)
 	 *	- with Port-Type = Sync (leased line up/down)
 	 */
@@ -442,26 +466,25 @@ int rad_accounting_new(AUTH_REQ *authreq, int dowtmp)
 	/*
 	 *	Enter into the radutmp file.
 	 */
-	if ((fd = open(RADUTMP, O_RDWR|O_CREAT, 0644)) >= 0) {
+	if ((fd = open(radutmp_path, O_RDWR|O_CREAT, 0644)) >= 0) {
 		NAS_PORT *cache;
 		int r;
 
 		/*
-		 *	Lock the utmp file, prefer lockf() over flock().
+		 *	Lock the utmp file.
 		 */
-#if defined(F_LOCK) && !defined(BSD)
-		(void)lockf(fd, F_LOCK, 0);
-#else
-		(void)flock(fd, LOCK_EX);
-#endif
+		lockfd(fd, 1);
+
 	 	/*
 		 *	Find the entry for this NAS / portno combination.
 		 */
-		if ((cache = nas_port_find(ut.nas_address, ut.nas_port)) != NULL)
+		off = 0;
+		if ((cache = nas_port_find(ut.nas_address, ut.nas_port)) != NULL) {
 			lseek(fd, (off_t)cache->offset, SEEK_SET);
+			off = (off_t)cache->offset;
+		}
 
 		r = 0;
-		off = 0;
 		while (read(fd, &u, sizeof(u)) == sizeof(u)) {
 			off += sizeof(u);
 			if (u.nas_address != ut.nas_address ||
@@ -559,7 +582,7 @@ int rad_accounting_new(AUTH_REQ *authreq, int dowtmp)
 		}
 		close(fd);
 	} else {
-		log(L_ERR, "Accounting: %s: %s", RADUTMP, strerror(errno));
+		log(L_ERR, "Accounting: %s: %s", radutmp_path, strerror(errno));
 		ret = -1;
 	}
 
@@ -574,7 +597,7 @@ int rad_accounting_new(AUTH_REQ *authreq, int dowtmp)
 	/*
 	 *	Write a RADIUS wtmp log file.
 	 */
-	if (dowtmp && (fp = fopen(RADWTMP, "a")) != NULL) {
+	if (dowtmp && (fp = fopen(radwtmp_path, "a")) != NULL) {
 		ret = 0;
 		fwrite(&wt, sizeof(wt), 1, fp);
 		fclose(fp);
@@ -584,21 +607,96 @@ int rad_accounting_new(AUTH_REQ *authreq, int dowtmp)
 	return ret;
 }
 
-
-
-int rad_accounting_orig(AUTH_REQ *authreq, int authtype, char *f)
+static int radius_xlate2(char *str, char *buf, int bufsz, AUTH_REQ *authreq)
 {
-	FILE		*outfd;
-	char		nasname[128];
-	char		buffer[512];
-	char		*s;
 	VALUE_PAIR	*pair;
-	UINT4		nas;
 	NAS		*cl;
-	int		flat = 0;
-	long		curtime;
-	int		ret = 0;
+	UINT4		nas;
+	char		rep[256];
+	char		*p, *s;
+	int		len, i = 0;
+
+	bufsz--;
+
+	for (p = str; *p; p++) {
+		if (i >= bufsz) break;
+		if (*p != '%') {
+			buf[i++] = *p;
+			continue;
+		}
+		if (*++p == 0) break;
+		switch(*p) {
+			case '%':
+				rep[0] = *p;
+				rep[1] = 0;
+				break;
+			case 'N':
+				/*
+				 *	Find IP address in order from:
+				 *	- remote proxy
+				 *	- NAS-IP-Address
+				 *	- Client-IP-Address
+				 *
+				 *	Then translate it to a hostname
+				 *	trying in order:
+				 *	- short nasname
+				 *	- long nasname
+				 *	- DNS lookup.
+				 */
+				cl = NULL;
+				nas = authreq->ipaddr;
+				if ((pair = pairfind(authreq->request,
+				     PW_NAS_IP_ADDRESS)) != NULL)
+					nas = pair->lvalue;
+
+				/*
+				 *	Reply from remote radius server:
+				 *	use server_ipaddr
+				 */
+				if (authreq->server_ipaddr)
+					nas = authreq->server_ipaddr;
+
+				if ((cl = nas_find(nas)) != NULL) {
+					/* {short,long}name are < 128 bytes */
+					if (cl->shortname[0])
+						strcpy(rep, cl->shortname);
+					else
+						strcpy(rep, cl->longname);
+				}
+
+				if (cl == NULL) {
+					s = ip_hostname(nas);
+					strNcpy(rep, s, sizeof(rep));
+				}
+				break;
+			default:
+				rep[0] = '%';
+				rep[1] = *p;
+				rep[2] = 0;
+				break;
+		}
+		len = strlen(rep);
+		if (i + len >= bufsz) break;
+		strcpy(buf + i, rep);
+		i += len;
+	}
+	buf[i++] = 0;
+
+	return 0;
+}
+
+int rad_accounting_detail(AUTH_REQ *authreq, int authtype, char *f)
+{
+	VALUE_PAIR	*pair;
+	FILE		*outfd;
 	struct stat	st;
+	struct timeval	tv;
+	time_t		curtime;
+	char		detail[512];
+	char		ipaddr[16];
+	char		*ptr;
+	char		*s;
+	int		len, ret = 0;
 
 	/*
 	 *	See if we have an accounting directory. If not,
@@ -607,74 +705,52 @@ int rad_accounting_orig(AUTH_REQ *authreq, int authtype, char *f)
 	if (stat(radacct_dir, &st) < 0)
 		return 0;
 	curtime = time(0);
-	nasname[0] = 0;
 
 	/*
-	 *	If radacct_dir ends in '/', assume we want a flat
-	 *	accouting directory (no subdirs, one detail file).
+	 *	Calculate detail file location. See if we need to
+	 *	create a subdirectory.
 	 */
-	if ((s = strrchr(radacct_dir, '/')) != NULL && s[1] == 0)
-		flat = 1;
-
-	if (!flat) {
-		/*
-		 *	Find out the name of this terminal server. We try
-		 *	to find the PW_NAS_IP_ADDRESS in the naslist file.
-		 *	If that fails, we look for the originating address.
-		 *	Only if that fails we resort to a name lookup.
-		 */
-		cl = NULL;
-		nas = authreq->ipaddr;
-		if ((pair = pairfind(authreq->request, PW_NAS_IP_ADDRESS)) != NULL)
-			nas = pair->lvalue;
-
-		/* Reply from remote radius server: use server_ipaddr */
-		if (authreq->server_ipaddr)
-			nas = authreq->server_ipaddr;
-
-		if ((cl = nas_find(nas)) != NULL) {
-			/* shortname and longname are < 128 bytes */
-			if (cl->shortname[0])
-				strcpy(nasname, cl->shortname);
-			else
-				strcpy(nasname, cl->longname);
-		}
-
-		if (cl == NULL) {
-			s = ip_hostname(nas);
-			if (strlen(s) >= sizeof(nasname) || strchr(s, '/'))
-				return -1;
-			strcpy(nasname, s);
-		}
-
-		/*
-		 *	Create a directory for this nas.
-		 */
-		sprintf(buffer, "%.250s/%.250s", radacct_dir, nasname);
-		(void) mkdir(buffer, 0755);
+	len = strlen(radacct_dir) + 1;
+	if (len >= sizeof(detail))
+		return -1;
+	strcpy(detail, radacct_dir);
+	strcat(detail, "/");
+	ptr = detail + len;
+	radius_xlate2(f, ptr, sizeof(detail) - len, authreq);
+	if ((s = strrchr(detail, '/')) != NULL && s > ptr) {
+		*s = 0;
+		(void) mkdir(detail, 0755);
+		*s = '/';
 	}
 
 	/*
-	 *	Write Detail file.
+	 *	Open detail file, lock it, and write the details.
 	 */
-	sprintf(buffer, "%.200s/%.200s%s%.100s", radacct_dir,
-		nasname, nasname[0] ? "/" : "", f ? f : "detail");
-	if ((outfd = fopen(buffer, "a")) == NULL) {
-		log(L_ERR, "Acct: Couldn't open file %s", buffer);
-		ret = -1;
-	} else {
+	do {
+		if ((outfd = fopen(detail, "a")) == NULL) {
+			log(L_ERR, "Acct: Couldn't open file %s", detail);
+			ret = -1;
+			break;
+		}
+
+		/*
+		 *	Try to lock the detail file, reopen it
+		 *	if it fails.
+		 */
+		if (lockfd(fileno(outfd), 2) < 0) {
+			tv.tv_sec = 0;
+			tv.tv_usec = 25000;
+			select(0, NULL, NULL, NULL, &tv);
+			fclose(outfd);
+			continue;
+		}
 
-#if defined(F_LOCK) && !defined(BSD)
-		(void)lockf(fileno(outfd), F_LOCK, 0);
-#else
-		(void)flock(fileno(outfd), LOCK_EX);
-#endif
 		/* Post a timestamp */
 		fputs(ctime(&curtime), outfd);
 
 		/* Write each attribute/value to the log file */
 		pair = authreq->request;
-		while(pair != (VALUE_PAIR *)NULL) {
+		while (pair != NULL) {
 			if (pair->attribute != PW_PASSWORD) {
 				fputs("\t", outfd);
 				fprint_attr_val(outfd, pair);
@@ -686,34 +762,40 @@ int rad_accounting_orig(AUTH_REQ *authreq, int authtype, char *f)
 		/*
 		 *	Add non-protocol attibutes.
 		 */
-		fprintf(outfd, "\tTimestamp = %ld\n", curtime);
-		switch(authtype) {
-		    case 0:
-			fputs("\tRequest-Authenticator = Verified\n", outfd);
-			break;
-		    case 1:
-			fputs("\tRequest-Authenticator = None\n", outfd);
-			break;
-		    case 2:
-			fputs("\tRequest-Authenticator = Unverified\n", outfd);
-			break;
-		    default:
-			break;
+		if (authreq->server_ipaddr) {
+			ipaddr2str(ipaddr, authreq->server_ipaddr);
+			fprintf(outfd, "\tCistron-Proxied-To = %s\n", ipaddr);
 		}
+		fprintf(outfd, "\tTimestamp = %ld\n", curtime);
+		if (authtype >= 0 && authtype <= 2)
+			fprintf(outfd, "\tRequest-Authenticator = %s\n",
+				detail_auth[authtype]);
 		fputs("\n", outfd);
 		fclose(outfd);
-	}
+	} while(0);
 
 	return ret;
 }
 
 
+/*
+ *	Add a detail file.
+ */
+int rad_add_detail(char *fn)
+{
+	if (detail_fn_count >= NR_DETAIL_FN)
+		return -1;
+	detail_fn[detail_fn_count++] = strdup(fn);
+	return 0;
+}
+
+
 /*
  *	rad_accounting: call both the old and new style accounting functions.
  */
 int rad_accounting(AUTH_REQ *authreq, int activefd)
 {
-	int reply = 0;
+	int i, reply = 0;
 	int auth;
 	char pw_digest[16];
 
@@ -730,22 +812,28 @@ int rad_accounting(AUTH_REQ *authreq, int activefd)
 
 	if (log_stripped_names) {
 		/*
-		 *	rad_accounting_new strips authreq for us. If
+		 *	rad_accounting_xtmp strips authreq for us. If
 		 *	we run it first, the stripped info will also
-		 *	get into the "details" file.
+		 *	get into the "detail" file.
 		 */
-		if (rad_accounting_new(authreq, use_wtmp) == 0)
-			reply = 1;
-		if (rad_accounting_orig(authreq, auth, NULL) == 0)
+		if (rad_accounting_radxtmp(authreq, use_wtmp) == 0)
 			reply = 1;
+		for (i = 0; i < 1 || i < detail_fn_count; i++) {
+			if (rad_accounting_detail(authreq, auth,
+			    detail_fn[i]) == 0)
+				reply = 1;
+		}
 	} else {
 		/*
 		 *	First log into the details file, before the
-		 *	username gets stripped by rad_accounting_new.
+		 *	username gets stripped by rad_accounting_xtmp.
 		 */
-		if (rad_accounting_orig(authreq, auth, NULL) == 0)
-			reply = 1;
-		if (rad_accounting_new(authreq, use_wtmp) == 0)
+		for (i = 0; i < 1 || i < detail_fn_count; i++) {
+			if (rad_accounting_detail(authreq, auth,
+			    detail_fn[i]) == 0)
+				reply = 1;
+		}
+		if (rad_accounting_radxtmp(authreq, use_wtmp) == 0)
 			reply = 1;
 	}
 
@@ -866,22 +954,24 @@ static int rad_check_ts(struct radutmp *ut)
  *
  *	Returns: 0 == OK, 1 == double logins, 2 == multilink attempt
  */
-int rad_check_multi(char *name, VALUE_PAIR *request, int maxsimul)
+int rad_check_multi(char *name, VALUE_PAIR *request, VALUE_PAIR *reply,
+	int maxsimul)
 {
+	VALUE_PAIR	*vp;
 	FILE		*wfp;
-	int		fd;
-	int		count;
-	struct radutmp	u;
+	struct radutmp	u, u2;
 	struct utmp	wt;
-	VALUE_PAIR	*fra;
-	int		mpp = 1;
-	UINT4		ipno = 0;
+	char		caller_id[16];
+	int		multi_count = 0;
+	int		portlimit;
+	int		count;
+	int		fd;
 
-	if ((fd = open(RADUTMP, O_CREAT|O_RDWR, 0644)) < 0)
+	if ((fd = open(radutmp_path, O_CREAT|O_RDWR, 0644)) < 0)
 		return 0;
 
 	/*
-	 *	We don't lock in the first pass.
+	 *	First pass: just read the radutmp file.
 	 */
 	count = 0;
 	while(read(fd, &u, sizeof(u)) == sizeof(u))
@@ -898,17 +988,12 @@ int rad_check_multi(char *name, VALUE_PAIR *request, int maxsimul)
 	/*
 	 *	Setup some stuff, like for MPP detection.
 	 */
-	if ((fra = pairfind(request, PW_FRAMED_IP_ADDRESS)) != NULL)
-		ipno = htonl(fra->lvalue);
-
-	/*
-	 *	lockf() the file while reading/writing.
-	 */
-#if defined(F_LOCK) && !defined(BSD)
-		(void)lockf(fd, F_LOCK, 0);
-#else
-		(void)flock(fd, LOCK_EX);
-#endif
+	caller_id[0] = 0;
+	portlimit = 1;
+	if ((vp = pairfind(request, PW_CALLING_STATION_ID)) != NULL)
+		strNcpy(caller_id, vp->strvalue, sizeof(caller_id));
+	if ((vp = pairfind(reply, PW_PORT_LIMIT)) != NULL)
+		portlimit = vp->lvalue;
 
 	/*
 	 *	Allright, there are too many concurrent logins.
@@ -924,20 +1009,25 @@ int rad_check_multi(char *name, VALUE_PAIR *request, int maxsimul)
 				/*
 				 *	Does it look like a MPP attempt?
 				 */
-				if (strchr("SCPA", u.proto) &&
-				    ipno && u.framed_address == ipno)
-					mpp = 2;
+				if (caller_id[0] &&
+				    strcmp(caller_id, u.caller_id) == 0)
+					multi_count++;
 			}
 			else {
 				/*
 				 *	False record - zap it.
 				 */
-
+				lockfd(fd, 1);
 				lseek(fd, -(off_t)sizeof(u), SEEK_CUR);
-				u.type = P_IDLE;
-				write(fd, &u, sizeof(u));
+				read(fd, &u2, sizeof(u2));
+				if (!strncmp(u2.login, u.login, RUT_NAMESIZE)){
+					u.type = P_IDLE;
+					lseek(fd, -(off_t)sizeof(u), SEEK_CUR);
+					write(fd, &u, sizeof(u));
+				}
+				lockfd(fd, 0);
 
-				if ((wfp = fopen(RADWTMP, "a")) != NULL) {
+				if ((wfp = fopen(radwtmp_path, "a")) != NULL) {
 					make_wtmp(&u, &wt, PW_STATUS_STOP);
 					fwrite(&wt, sizeof(wt), 1, wfp);
 					fclose(wfp);
@@ -947,6 +1037,6 @@ int rad_check_multi(char *name, VALUE_PAIR *request, int maxsimul)
 	}
 	close(fd);
 
-	return (count < maxsimul) ? 0 : mpp;
+	return (count < maxsimul) ? 0 : (multi_count ? 2 : 1);
 }
 
