@@ -1,33 +1,11 @@
 /*
+ * attrprint.c	Functions to print A/V pairs.
  *
- *	RADIUS
- *	Remote Authentication Dial In User Service
- *
- *
- *	Livingston Enterprises, Inc.
- *	6920 Koll Center Parkway
- *	Pleasanton, CA   94566
- *
- *	Copyright 1992 Livingston Enterprises, Inc.
- *
- *	Permission to use, copy, modify, and distribute this software for any
- *	purpose and without fee is hereby granted, provided that this
- *	copyright and permission notice appear on all copies and supporting
- *	documentation, the name of Livingston Enterprises, Inc. not be used
- *	in advertising or publicity pertaining to distribution of the
- *	program without specific prior permission, and notice be given
- *	in supporting documentation that copying and distribution is by
- *	permission of Livingston Enterprises, Inc.   
- *
- *	Livingston Enterprises, Inc. makes no representations about
- *	the suitability of this software for any purpose.  It is
- *	provided "as is" without express or implied warranty.
- *
+ *		Copyright 1992 Livingston Enterprises Inc\
+ *		Copyright 1998-2001 Cistron Internet Services B.V.
  */
-
-char attrprint_sccsid[] =
-"@(#)attrprint.c	1.3 Copyright 1992 Livingston Enterprises Inc\n"
-"			2.1 Copyright 1998 Cistron Internet Services B.V.";
+char attrprint_rcsid[] =
+"$Id: attrprint.c,v 1.8 2001/11/14 16:15:00 miquels Exp $";
 
 #include	<sys/types.h>
 #include	<sys/socket.h>
@@ -70,6 +48,30 @@ void fprint_attr_list(FILE *fd, VALUE_PAIR *pair)
 	}
 }
 
+static void cprint(FILE *fp, int c)
+{
+	if (c == '\\' || c == '"' || c < 32 || c > 127) switch (c) {
+		case '\\':
+			fputs("\\\\", fp);
+			break;
+		case '"':
+			fputs("\\\"", fp);
+			break;
+		case '\r':
+			fputs("\\r", fp);
+			break;
+		case '\n':
+			fputs("\\n", fp);
+			break;
+		case '\t':
+			fputs("\\t", fp);
+			break;
+		default:
+			fprintf(fp, "\\%03o", c);
+			break;
+	} else
+		fputc(c, fp);
+}
 
 /*
  *	Write a printable version of the attribute-value
@@ -84,10 +86,21 @@ void fprint_attr_val(FILE *fd, VALUE_PAIR *pair)
 	UINT4		vendor;
 	int		i, left;
 
+	if (pair->flags.has_tag) {
+#ifdef MERIT_TAG_FORMAT
+		fprintf(fd, "%s = :%d:", pair->name,pair->flags.tag);
+#else
+		fprintf(fd, "%s:%d = ", pair->name,pair->flags.tag);
+#endif
+	}
+	else {
+		fprintf(fd, "%s = ",pair->name);
+	}
+
 	switch(pair->type) {
 
 	case PW_TYPE_STRING:
-		fprintf(fd, "%s = \"", pair->name);
+		fputc('"',fd);
 		ptr = (u_char *)pair->strvalue;
 		if (pair->attribute != PW_VENDOR_SPECIFIC) {
 			left = pair->length;
@@ -99,10 +112,7 @@ void fprint_attr_val(FILE *fd, VALUE_PAIR *pair)
 				 */
 				if (left == 0 && *ptr == 0)
 					break;
-				if(!(isprint(*ptr)))
-					fprintf(fd, "\\%03o", *ptr);
-				else
-					fputc(*ptr, fd);
+				cprint(fd, *ptr);
 				ptr++;
 			}
 			fputc('"', fd);
@@ -121,16 +131,26 @@ void fprint_attr_val(FILE *fd, VALUE_PAIR *pair)
 		fprintf(fd, "V%d", (int)ntohl(vendor));
 		left = pair->length - 4;
 		while (left >= 2) {
-			fprintf(fd, ":T%d:L%d:", ptr[0], ptr[1]);
-			left -= 2;
-			ptr += 2;
-			i = ptr[1] - 2;
+#ifdef ATTRIB_NMC
+			if (ntohl(vendor) == VENDORPEC_USR) {
+				int type;
+				if (left < 4) break;
+				memcpy(&type, ptr, 4);
+				left -= 4;
+				i = left;
+				fprintf(fd, ":T%d:L%d:",
+					(int)ntohl(type), i);
+				ptr += 4;
+			} else
+#endif
+			{
+				left -= 2;
+				i = ptr[1] - 2;
+				fprintf(fd, ":T%d:L%d:", ptr[0], i);
+				ptr += 2;
+			}
 			while (i > 0 && left > 0) {
-				if(!(isprint(*ptr)))
-					fprintf(fd, "\\%03o", *ptr);
-				else
-					fputc(*ptr, fd);
-				ptr++;
+				cprint(fd, *ptr++);
 				i--;
 				left--;
 			}
@@ -141,22 +161,29 @@ void fprint_attr_val(FILE *fd, VALUE_PAIR *pair)
 	case PW_TYPE_INTEGER:
 		dval = dict_valget(pair->lvalue, pair->name);
 		if(dval != (DICT_VALUE *)NULL) {
-			fprintf(fd, "%s = %s", pair->name, dval->name);
+			fprintf(fd, "%s", dval->name);
 		}
 		else {
-			fprintf(fd, "%s = %ld", pair->name, (long)pair->lvalue);
+			fprintf(fd, "%lu", (unsigned long)pair->lvalue);
 		}
 		break;
 
 	case PW_TYPE_IPADDR:
-		ipaddr2str(buffer, pair->lvalue);
-		fprintf(fd, "%s = %s", pair->name, buffer);
+		if (pair->lvalue == 0 && pair->strvalue[0] != 0) {
+			/* Why is this??? - EvB */
+			/* Because of the 192.168.1.1+ format -- miquels */
+			fprintf(fd, "%s", pair->strvalue);
+		}
+		else {
+			ipaddr2str(buffer, pair->lvalue);
+			fprintf(fd, "%s", buffer);
+		}
 		break;
 
 	case PW_TYPE_DATE:
 		strftime(buffer, sizeof(buffer), "%b %e %Y",
 					localtime((time_t *)&pair->lvalue));
-		fprintf(fd, "%s = \"%s\"", pair->name, buffer);
+		fprintf(fd, "\"%s\"", buffer);
 		break;
 
 	default:
