@@ -1,35 +1,13 @@
 /*
+ * radiusd.c	Main loop of the server.
  *
- *	RADIUS
- *	Remote Authentication Dial In User Service
- *
- *
- *	Livingston Enterprises, Inc.
- *	6920 Koll Center Parkway
- *	Pleasanton, CA   94566
- *
- *	Copyright 1992 Livingston Enterprises, Inc.
- *	Copyright 1998 Cistron Internet Services B.V.
- *
- *	Permission to use, copy, modify, and distribute this software for any
- *	purpose and without fee is hereby granted, provided that this
- *	copyright and permission notice appear on all copies and supporting
- *	documentation, the name of Livingston Enterprises, Inc. not be used
- *	in advertising or publicity pertaining to distribution of the
- *	program without specific prior permission, and notice be given
- *	in supporting documentation that copying and distribution is by
- *	permission of Livingston Enterprises, Inc.   
- *
- *	Livingston Enterprises, Inc. makes no representations about
- *	the suitability of this software for any purpose.  It is
- *	provided "as is" without express or implied warranty.
- *
+ *		Copyright 1992		Livingston Enterprises, Inc.
+ *		Copyright 1998-2001	Cistron Internet Services B.V.
  */
 
 /* don't look here for the version, run radiusd -v or look in version.c */
-char radiusd_sccsid[] =
-"@(#)radiusd.c	1.17 Copyright 1992 Livingston Enterprises Inc\n"
-"		1.84 Copyright 1999 Cistron Internet Services B.V.";
+char radiusd_rcsid[] =
+"$Id: radiusd.c,v 1.22 2001/11/10 12:41:39 miquels Exp $";
 
 #include	<sys/types.h>
 #include	<sys/socket.h>
@@ -59,18 +37,23 @@ char radiusd_sccsid[] =
  */
 char			*progname;
 char			*radius_dir;
-char			*radlog_dir;
+extern char		*radlog_dir;
+char			*radutmp_path;
+char			*radwtmp_path;
+#ifdef USE_SYSLOG
+extern int		syslog_facility;
+#endif
 char			*radacct_dir;
 int			log_stripped_names;
 int 			cache_passwd = 0;
-int			debug_flag;
 int			use_dbm = 0;
 int			use_wtmp = 1;
 UINT4			myip = 0;
 UINT4			warning_seconds;
-int			log_auth_detail = 0;
+char			*log_auth_detail;
 int			log_auth = 0;
-int			log_auth_pass  = 0;
+int			log_auth_badpass  = 0;
+int			log_auth_goodpass  = 0;
 int			auth_port;
 int			acct_port;
 
@@ -88,6 +71,7 @@ static int		spawn_flag;
 static int		acct_pid;
 static int		radius_pid;
 static int		need_reload = 0;
+static time_t		start_time;
 static AUTH_REQ		*first_request;
 
 
@@ -121,7 +105,7 @@ static void reread_config(int reload)
 	}
 
 	/* Initialize the dictionary */
-	if (dict_init(NULL) != 0)
+	if (dict_init(radius_dir, NULL) != 0)
 		res = -1;
 
 	/* Initialize Configuration Values */
@@ -164,11 +148,16 @@ int main(int argc, char **argv)
 	int			radius_port = 0;
 	int			check_config = 0;
 #ifdef RADIUS_PID
+	char			*pid_file = RADIUS_PID;
 	FILE			*fp;
 #endif
-#ifdef RLIMIT_CORE
-	struct rlimit rlim;
-#endif
+
+	/*
+	 *	Make sure we have stdin/stdout/stderr reserved.
+	 */
+	while((t = open("/dev/null", O_RDWR)) < 3 && t >= 0)
+		;
+	if (t >= 3) close(t);
 
 #ifdef OSFC2
 	set_auth_parameters(argc,argv);
@@ -184,6 +173,9 @@ int main(int argc, char **argv)
 	radacct_dir = RADACCT_DIR;
 	radlog_dir = RADLOG_DIR;
 	radius_dir = RADIUS_DIR;
+	radutmp_path = RADUTMP;
+	radwtmp_path = RADWTMP;
+	start_time = time(NULL);
 
 	signal(SIGHUP, sig_hup);
 	signal(SIGINT, sig_fatal);
@@ -207,19 +199,19 @@ int main(int argc, char **argv)
 	/*
 	 *	Process the options.
 	 */
-	while((argval = getopt(argc, argv, "ACSDa:ci:l:d:bfp:svwxyz")) != EOF) {
+	while((argval = getopt(argc, argv, "A:CSDF:P:W:Za:ci:l:d:bfp:g:su:vwxyz")) != EOF) {
 
 		switch(argval) {
 
 		case 'A':
-			log_auth_detail++;
+			log_auth_detail = optarg;
 			break;
 
 		case 'a':
 			radacct_dir = optarg;
 			break;
 		
-#ifdef USE_DBM
+#ifdef USE_DBX
 		case 'b':
 			use_dbm++;
 			break;
@@ -240,6 +232,10 @@ int main(int argc, char **argv)
 			radius_dir = optarg;
 			break;
 		
+		case 'F':
+			rad_add_detail(optarg);
+			break;
+
 		case 'f':
 			dontfork = 1;
 			break;
@@ -256,6 +252,19 @@ int main(int argc, char **argv)
 			radlog_dir = optarg;
 			break;
 		
+#ifdef USE_SYSLOG
+		case 'g':
+			syslog_facility = -1;
+			if (strncmp(optarg, "local", 5) == 0 && !optarg[6])
+				syslog_facility = optarg[5] - '0';
+			if (syslog_facility < 0 || syslog_facility > 7) {
+				fprintf(stderr,
+				"radiusd: unknown syslog facility: %s\n",
+				optarg);
+				exit(1);
+			}
+#endif
+
 		case 'S':
 			log_stripped_names++;
 			break;
@@ -264,14 +273,28 @@ int main(int argc, char **argv)
 			radius_port = atoi(optarg);
 			break;
 
+#ifdef RADIUS_PID
+		case 'P':
+			pid_file = optarg;
+			break;
+#endif
+
 		case 's':	/* Single process mode */
 			spawn_flag = 0;
 			break;
 
+		case 'u':
+			radutmp_path = optarg;
+			break;
+
 		case 'v':
 			version();
 			break;
 
+		case 'W':
+			radwtmp_path = optarg;
+			break;
+
 		case 'w':
 			use_wtmp = 0;
 			break;
@@ -282,10 +305,18 @@ int main(int argc, char **argv)
 		
 		case 'y':
 			log_auth = 1;
+			log_auth_goodpass = 0;
+			log_auth_badpass = 1;
 			break;
 
 		case 'z':
-			log_auth_pass = 1;
+			log_auth_badpass = 1;
+			log_auth_goodpass = 1;
+			break;
+
+		case 'Z':
+			log_auth_badpass = 0;
+			log_auth_goodpass = 0;
 			break;
 
 		default:
@@ -294,14 +325,17 @@ int main(int argc, char **argv)
 		}
 	}
 
+#ifdef USE_SYSLOG
+	if (strcmp(radlog_dir, "syslog") == 0 && syslog_facility < 0) {
+		fprintf(stderr,
+			"radiusd: -l syslog needs -g syslog_facility\n");
+		exit(1);
+	}
+#endif
+
 	if (check_config) {
 		debug_flag = 1;
-#if 0
-		fclose(stderr);
-		stderr = stdout;
-#else
 		dup2(1, 2);
-#endif
 		reread_config(0);
 		printf("Configuration OK\n");
 		return 0;
@@ -371,7 +405,7 @@ int main(int argc, char **argv)
 	/*
 	 *	Disconnect from session
 	 */
-	if(debug_flag == 0 && dontfork == 0) {
+	if (debug_flag == 0 && dontfork == 0) {
 		pid = fork();
 		if(pid < 0) {
 			log(L_ERR|L_CONS, "Couldn't fork");
@@ -380,13 +414,20 @@ int main(int argc, char **argv)
 		if(pid > 0) {
 			exit(0);
 		}
-#if defined(__linux__) || defined(__svr4__) || defined(__hpux__)
+		close(0);
+		close(1);
+		close(2);
+		(void)open("/dev/null", O_RDWR);
+		dup(0);
+		dup(0);
+#if defined(__linux__) || defined(__svr4__) || \
+    defined(__hpux__)  || defined(__FreeBSD__)
 		setsid();
 #endif
 	}
 	radius_pid = getpid();
 #ifdef RADIUS_PID
-	if ((fp = fopen(RADIUS_PID, "w")) != NULL) {
+	if ((fp = fopen(pid_file, "w")) != NULL) {
 		fprintf(fp, "%d\n", radius_pid);
 		fclose(fp);
 	}
@@ -402,7 +443,7 @@ int main(int argc, char **argv)
 	/*
 	 *	Open system console as stderr
 	 */
-	if (!debug_flag) {
+	if (!debug_flag && strcmp(radlog_dir, "stderr") != 0) {
 		t = open("/dev/console", O_WRONLY | O_NOCTTY);
 		if (t != 2) {
 			dup2(t, 2);
@@ -500,6 +541,32 @@ int main(int argc, char **argv)
 }
 
 
+/*
+ *	Process and reply to a server-status request.
+ *	Like rad_authenticate and rad_accounting this should
+ *	live in it's own file but it's so small we don't bother.
+ */
+int rad_status_server(AUTH_REQ *authreq, int activefd)
+{
+	char		reply_msg[64];
+	time_t		t;
+
+	/*
+	 *	Reply with an ACK. We might want to add some more
+	 *	interesting reply attributes, such as server uptime.
+	 */
+	t = time(NULL) - start_time;
+	sprintf(reply_msg, "Cistron Radius up %d day%s, %02d:%02d",
+		(int)(t / 86400), (t / 86400) == 1 ? "" : "s",
+		(int)((t / 3600) % 24), (int)(t / 60) % 60);
+	rad_send_reply(PW_AUTHENTICATION_ACK, authreq,
+			NULL, reply_msg, activefd);
+	authfree(authreq);
+
+	return 0;
+}
+
+
 /*
  *	Respond to supported requests:
  *
@@ -514,17 +581,40 @@ int main(int argc, char **argv)
  *		PW_ACCOUNTING_RESPONSE - Reply from a remote Radius server.
  *				Relay reply back to original NAS.
  *
+ *		PW_SERVER_STATUS - Usually an "alive" check.
+ *
  */
 int radrespond(AUTH_REQ *authreq, int activefd)
 {
-	int dospawn;
 	FUNP fun;
 	VALUE_PAIR *namepair;
+	RADCLIENT *cl;
+	int dospawn;
 	int e;
 
 	dospawn = 0;
 	fun = NULL;
 
+	/*
+	 *      See if we know this client.
+	 */
+        if ((cl = client_find(authreq->ipaddr)) == NULL) {
+                log(L_ERR, "packet from unknown client/host: %s",
+                        client_name(authreq->ipaddr));
+                return -1;
+        }
+	strNcpy(authreq->secret, cl->secret, sizeof(authreq->secret));
+
+#ifdef USE_DBX
+	/*
+	 *	When using the DBM code, read or re-read the
+	 *	DBM database. We do this in the parent, before
+	 *	we fork, so that we don't have to reopen the
+	 *	database for every request.
+	 */
+	if (use_dbm) rad_dbm_open();
+#endif
+
 	/*
 	 *	First, see if we need to proxy this request.
 	 */
@@ -579,7 +669,10 @@ int radrespond(AUTH_REQ *authreq, int activefd)
 		 */
 		/* rad_passchange(authreq, activefd); */
 		break;
-	
+
+	case PW_STATUS_SERVER:
+		fun = rad_status_server;
+		break;
 
 	default:
 		break;
@@ -611,6 +704,7 @@ static void rad_spawn_child(AUTH_REQ *authreq, int activefd, FUNP fun)
 {
 	AUTH_REQ	*curreq;
 	AUTH_REQ	*prevreq;
+	VALUE_PAIR	*vp;
 	UINT4		curtime;
 	int		request_count;
 	int		child_pid;
@@ -643,40 +737,23 @@ static void rad_spawn_child(AUTH_REQ *authreq, int activefd, FUNP fun)
 				curreq = prevreq->next;
 			}
 		} else if (curreq->ipaddr == authreq->ipaddr &&
+			   curreq->udp_port == authreq->udp_port &&
 			   curreq->id == authreq->id) {
 			/*
-			 *	Compare the request vectors to see
-			 *	if it really is the same request.
+			 *	This is a duplicate request - we are still
+			 *	processing it, or sent a reply less than
+			 *	3 seconds ago. Just drop it.
 			 */
-			if (memcmp(curreq->vector, authreq->vector, 16) == 0) {
-				/*
-				 * This is a duplicate request - just drop it
-				 */
-				log(L_ERR,
-				"Dropping duplicate authentication packet"
-				" from client %s - ID: %d",
-				client_name(authreq->ipaddr), authreq->id);
-				authfree(authreq);
-				request_list_busy = 0;
-				sig_cleanup(SIGCHLD);
+			vp = pairfind(authreq->request, PW_USER_NAME);
+			log(L_ERR,
+			"Duplicate authentication packet: [%s] (%s/ID:%d)",
+				vp ? vp->strvalue : "",
+				auth_name(authreq, 0), authreq->id);
+			authfree(authreq);
+			request_list_busy = 0;
+			sig_cleanup(SIGCHLD);
 
-				return;
-			}
-			/*
-			 *	If the old request was completed,
-			 *	delete it right now.
-			 */
-			if (curreq->child_pid == -1) {
-				curreq->timestamp = curtime - CLEANUP_DELAY;
-				continue;
-			}
-
-			/*
-			 *	Not completed yet, do nothing special.
-			 */
-			prevreq = curreq;
-			curreq = curreq->next;
-			request_count++;
+			return;
 		} else {
 			if (curreq->timestamp + MAX_REQUEST_TIME <= curtime &&
 			    curreq->child_pid != -1) {
@@ -701,13 +778,15 @@ static void rad_spawn_child(AUTH_REQ *authreq, int activefd, FUNP fun)
 	 *	This is a new request
 	 */
 	if (request_count > MAX_REQUESTS) {
-		log(L_ERR, "Dropping request (too many): from client %s - ID: %d",
-				client_name(authreq->ipaddr), authreq->id);
+		log(L_ERR,
+		"Dropping packet (too many): from host %s:%d - ID: %d",
+			client_name(authreq->ipaddr), authreq->udp_port,
+			authreq->id);
 		authfree(authreq);
 
 		request_list_busy = 0;
 		sig_cleanup(SIGCHLD);
-				
+
 		return;
 	}
 
@@ -807,10 +886,10 @@ void sig_cleanup(int sig)
 static void usage(void)
 {
 	fprintf(stderr,
-#ifdef USE_DBM
-		"Usage: %s [-a acct_dir] [-d db_dir] [-l logdir] [-bcsxyz]\n",
+#ifdef USE_DBX
+		"Usage: %s [-a acct_dir] [-d db_dir] [-l logdir] [-bcsxYyz]\n",
 #else
-		"Usage: %s [-a acct_dir] [-d db_dir] [-l logdir] [-csxyz]\n",
+		"Usage: %s [-a acct_dir] [-d db_dir] [-l logdir] [-csxYyz]\n",
 #endif
 		progname);
 	exit(1);
@@ -831,12 +910,11 @@ int config_init()
 {
 	DICT_VALUE	*dval;
 
-	if((dval = dict_valfind("Password-Warning")) == (DICT_VALUE *)NULL) {
-		warning_seconds = (UINT4)0;
-	}
-	else {
+	warning_seconds = 0;
+	dval = dict_valfind("Password-Warning", "Server-Config");
+	if (dval != NULL)
 		warning_seconds = dval->value * (UINT4)SECONDS_PER_DAY;
-	}
+
 	return 0;
 }
 
