@@ -1,34 +1,15 @@
 /*
  *
- *	RADIUS
- *	Remote Authentication Dial In User Service
- *
- *
- *	Livingston Enterprises, Inc.
- *	6920 Koll Center Parkway
- *	Pleasanton, CA   94566
- *
- *	Copyright 1992 Livingston Enterprises, Inc.
- *	Copyright 1997 Cistron Internet Services B.V.
- *
- *	Permission to use, copy, modify, and distribute this software for any
- *	purpose and without fee is hereby granted, provided that this
- *	copyright and permission notice appear on all copies and supporting
- *	documentation, the name of Livingston Enterprises, Inc. not be used
- *	in advertising or publicity pertaining to distribution of the
- *	program without specific prior permission, and notice be given
- *	in supporting documentation that copying and distribution is by
- *	permission of Livingston Enterprises, Inc.   
- *
- *	Livingston Enterprises, Inc. makes no representations about
- *	the suitability of this software for any purpose.  It is
- *	provided "as is" without express or implied warranty.
+ * util.c	Several miscellaneous functions. For example to
+ *		deal with hostnames and IP addresses (in host-order,
+ *		how braindead can you get...)
  *
+ *		Copyright 1992		Livingston Enterprises, Inc.
+ *		Copyright 1997-2001	Cistron Internet Services B.V.
  */
 
-char util_sccsid[] =
-"@(#)util.c	1.5 Copyright 1992 Livingston Enterprises Inc\n"
-"		2.1 Copyright 1997 Cistron Internet Services B.V.";
+char util_rcsid[] =
+"$Id: util.c,v 1.8 2001/02/09 00:18:28 miquels Exp $";
 
 #include	<sys/types.h>
 #include	<sys/socket.h>
@@ -41,9 +22,12 @@ char util_sccsid[] =
 #include	<pwd.h>
 #include	<time.h>
 #include	<ctype.h>
+#include	<signal.h>
 
 #include	"radiusd.h"
 
+int	use_dns = 1;
+
 /*
  *	Return a printable host name (or IP address in dot notation)
  *	for the supplied IP address.
@@ -55,8 +39,12 @@ char * ip_hostname(UINT4 ipaddr)
 	UINT4		n_ipaddr;
 
 	n_ipaddr = htonl(ipaddr);
-	hp = gethostbyaddr((char *)&n_ipaddr, sizeof (struct in_addr), AF_INET);
-	if (hp == 0) {
+	if (use_dns)
+		hp = gethostbyaddr((char *)&n_ipaddr,
+			sizeof (struct in_addr), AF_INET);
+	else
+		hp = NULL;
+	if (hp == NULL) {
 		ipaddr2str(hstname, ipaddr);
 		return(hstname);
 	}
@@ -201,67 +189,53 @@ struct passwd *rad_getpwnam(char *name)
 	return lastpwd;
 }
 
-
 /*
- *	Release the memory used by a list of attribute-value
- *	pairs.
+ *	Free an AUTHREQ struct.
  */
-void pairfree(VALUE_PAIR *pair)
+void authfree(AUTH_REQ *authreq)
 {
-	VALUE_PAIR	*next;
-
-	while(pair != NULL) {
-		next = pair->next;
-		free(pair);
-		pair = next;
-	}
+	pairfree(authreq->request);
+	pairfree(authreq->proxy_pairs);
+	pairfree(authreq->server_reply);
+	memset(authreq, 0, sizeof(AUTH_REQ));
+	free(authreq);
 }
 
+#if (defined (sun) && defined(__svr4__)) || defined(__hpux__) || defined(aix)
+/*
+ *	The signal() function in Solaris 2.5.1 sets SA_NODEFER in
+ *	sa_flags, which causes grief if signal() is called in the
+ *	handler before the cause of the signal has been cleared.
+ *	(Infinite recursion).
+ */
+void (*sun_signal(int signo, void (*func)(int)))(int)
+{
+	struct sigaction act, oact;
+
+	act.sa_handler = func;
+	sigemptyset(&act.sa_mask);
+	act.sa_flags = 0;
+#ifdef  SA_INTERRUPT		/* SunOS */
+	act.sa_flags |= SA_INTERRUPT;
+#endif
+	if (sigaction(signo, &act, &oact) < 0)
+		return SIG_ERR;
+	return oact.sa_handler;
+}
+#endif
 
 /*
- *	Find the pair with the mathing attribute
+ *	Like strncpy, but makes sure that the string
+ *	always ends with a trailing \0
  */
-VALUE_PAIR * pairfind(VALUE_PAIR *first, int attr)
+char *strNcpy(char *dest, char *src, int n)
 {
-	while(first && first->attribute != attr)
-		first = first->next;
-	return first;
+	if (n > 0)
+		strncpy(dest, src, n);
+	else
+		n = 1;
+	dest[n - 1] = 0;
+
+	return dest;
 }
 
-
-/*
- *	Delete the pair(s) with the mathing attribute
- */
-void pairdelete(VALUE_PAIR **first, int attr)
-{
-	VALUE_PAIR *i, *next, *last = NULL;
-
-	for(i = *first; i; i = next) {
-		next = i->next;
-		if (i->attribute == attr) {
-			if (last)
-				last->next = next;
-			else
-				*first = next;
-			free(i);
-		} else
-			last = i;
-	}
-}
-
-/*
- *	Add a pair at the end of a VALUE_PAIR list.
- */
-void pairadd(VALUE_PAIR **first, VALUE_PAIR *new)
-{
-	VALUE_PAIR *i;
-
-	new->next = NULL;
-	if (*first == NULL) {
-		*first = new;
-		return;
-	}
-	for(i = *first; i->next; i = i->next)
-		;
-	i->next = new;
-}
